IDAStar = {}

function IDAStar:new()
	local idastar = {}
	setmetatable(idastar, { __index = IDAStar })

	return idastar
end

function IDAStar:search(state, depth, path)
	if depth == 0 then
		if self:solved(state) then
			coroutine.yield(path)
		end

		return
	end

	if not self:prune(state, depth) then
		for move, next_state in self:next_states(state, path) do
			table.insert(path, move)
			self:search(next_state, depth - 1, path)
			table.remove(path)
		end
	end
end

function IDAStar:solutions(state, max_length)
	max_length = max_length or 1e6

	if self.preprocess_state then
		state = self:preprocess_state(state)
	end

	return coroutine.wrap(function()
		for depth = 0, max_length do
			self:search(state, depth, {})
		end
	end)
end

return IDAStar
