State = {}

State.metatable = { __index = State }

function State:new(pCorners, oCorners, pEdges, oEdges)
    local state = {
        pCorners = pCorners,
        oCorners = oCorners,
        pEdges = pEdges,
        oEdges = oEdges,
    }
    setmetatable(state, self.metatable)

    return state
end

function State.from_scramble(scramble)
    local result = State.id
    for _, move_name in ipairs(scramble) do
        result = result:mul(State.moves[move_name])
    end

    return result
end

function State:mul(state)
    local pCorners = {}
    local oCorners = {}
    for i = 1, 8 do
        pCorners[i] = self.pCorners[state.pCorners[i]]
        oCorners[i] = (self.oCorners[state.pCorners[i]] + state.oCorners[i]) % 3
    end

    local pEdges = {}
    local oEdges = {}
    for i = 1, 12 do
        pEdges[i] = self.pEdges[state.pEdges[i]]
        oEdges[i] = (self.oEdges[state.pEdges[i]] + state.oEdges[i]) % 2
    end

    return State:new(pCorners, oCorners, pEdges, oEdges)
end

function State:pow(n)
    local state = State.id
    for i = 1, n do
        state = state:mul(self)
    end

    return state
end

function State:inverse()
    local pCorners = {}
    local oCorners = {}
    for i = 1, 8 do
        pCorners[self.pCorners[i]] = i
        oCorners[self.pCorners[i]] = (3 - self.oCorners[i]) % 3
    end

    local pEdges = {}
    local oEdges = {}
    for i = 1, 12 do
        pEdges[self.pEdges[i]] = i
        oEdges[self.pEdges[i]] = (2 - self.oEdges[i]) % 2
    end

    return State:new(pCorners, oCorners, pEdges, oEdges)
end

function State:equals(state)
    for i = 1, 8 do
        if self.pCorners[i] ~= state.pCorners[i] or
           self.oCorners[i] ~= state.oCorners[i] then
            return false
        end
    end

    for i = 1, 12 do
        if self.pEdges[i] ~= state.pEdges[i] or
           self.oEdges[i] ~= state.oEdges[i] then
            return false
        end
    end

    return true
end
State.metatable.__eq = State.equals

State.id = State:new({ 1, 2, 3, 4, 5, 6, 7, 8 }, { 0, 0, 0, 0, 0, 0, 0, 0 }, { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 })

local moveU = State:new({ 4, 1, 2, 3, 5, 6, 7, 8 }, { 0, 0, 0, 0, 0, 0, 0, 0 }, { 1, 2, 3, 4, 8, 5, 6, 7, 9, 10, 11, 12 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 })
local moveD = State:new({ 1, 2, 3, 4, 6, 7, 8, 5 }, { 0, 0, 0, 0, 0, 0, 0, 0 }, { 1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12, 9 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 })
local moveL = State:new({ 5, 2, 3, 1, 8, 6, 7, 4 }, { 2, 0, 0, 1, 1, 0, 0, 2 }, { 12, 2, 3, 8, 5, 6, 7, 1, 9, 10, 11, 4 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 })
local moveR = State:new({ 1, 3, 7, 4, 5, 2, 6, 8 }, { 0, 1, 2, 0, 0, 2, 1, 0 }, { 1, 6, 10, 4, 5, 3, 7, 8, 9, 2, 11, 12 }, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 })
local moveF = State:new({ 1, 2, 4, 8, 5, 6, 3, 7 }, { 0, 0, 1, 2, 0, 0, 2, 1 }, { 1, 2, 7, 11, 5, 6, 4, 8, 9, 10, 3, 12 }, { 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0 })
local moveB = State:new({ 2, 6, 3, 4, 1, 5, 7, 8 }, { 1, 2, 0, 0, 2, 1, 0, 0 }, { 5, 9, 3, 4, 2, 6, 7, 8, 1, 10, 11, 12 }, { 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 })

State.moves = {
    ["U"] = moveU, ["U2"] = moveU:pow(2), ["U'"] = moveU:pow(3),
    ["D"] = moveD, ["D2"] = moveD:pow(2), ["D'"] = moveD:pow(3),
    ["L"] = moveL, ["L2"] = moveL:pow(2), ["L'"] = moveL:pow(3),
    ["R"] = moveR, ["R2"] = moveR:pow(2), ["R'"] = moveR:pow(3),
    ["F"] = moveF, ["F2"] = moveF:pow(2), ["F'"] = moveF:pow(3),
    ["B"] = moveB, ["B2"] = moveB:pow(2), ["B'"] = moveB:pow(3),
}

return State
