package com.puzzletimer.solvers;

import java.util.HashMap;

public class State {
    public int[] pCorners;
    public int[] oCorners;
    public int[] pEdges;
    public int[] oEdges;

    public State(int[] pCorners, int[] oCorners, int[] pEdges, int[] oEdges) {
        this.pCorners = pCorners;
        this.oCorners = oCorners;
        this.pEdges = pEdges;
        this.oEdges = oEdges;
    }

    public State mul(State s) {
        int[] pCorners = new int[8];
        int[] oCorners = new int[8];

        for (int i = 0; i < 8; i++) {
            pCorners[i] = this.pCorners[s.pCorners[i]];
            oCorners[i] = (this.oCorners[s.pCorners[i]] + s.oCorners[i]) % 3;
        }

        int[] pEdges = new int[12];
        int[] oEdges = new int[12];

        for (int i = 0; i < 12; i++) {
            pEdges[i] = this.pEdges[s.pEdges[i]];
            oEdges[i] = (this.oEdges[s.pEdges[i]] + s.oEdges[i]) % 2;
        }

        return new State(pCorners, oCorners, pEdges, oEdges);
    }

    public State pow(int n) {
        State state = State.id;
        for (int i = 0; i < n; i++) {
            state = state.mul(this);
        }

        return state;
    }

    public State apply(String[] sequence) {
        State state = this;
        for (String move : sequence) {
            state = state.mul(moves.get(move));
        }

        return state;
    }

    public static State id;

    static {
        id = new State(
            new int[] { 0, 1, 2, 3, 4, 5, 6, 7 },
            new int[] { 0, 0, 0, 0, 0, 0, 0, 0 },
            new int[] { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 },
            new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0,  0 });
    }

    public static HashMap<String, State> moves;

    static {
        State moveU = new State(new int[] { 3, 0, 1, 2, 4, 5, 6, 7 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0 }, new int[] { 0, 1, 2, 3, 7, 4, 5, 6, 8, 9, 10, 11 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 });
        State moveD = new State(new int[] { 0, 1, 2, 3, 5, 6, 7, 4 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0 }, new int[] { 0, 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 8 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 });
        State moveL = new State(new int[] { 4, 1, 2, 0, 7, 5, 6, 3 }, new int[] { 2, 0, 0, 1, 1, 0, 0, 2 }, new int[] { 11, 1, 2, 7, 4, 5, 6, 0, 8, 9, 10, 3 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 });
        State moveR = new State(new int[] { 0, 2, 6, 3, 4, 1, 5, 7 }, new int[] { 0, 1, 2, 0, 0, 2, 1, 0 }, new int[] { 0, 5, 9, 3, 4, 2, 6, 7, 8, 1, 10, 11 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 });
        State moveF = new State(new int[] { 0, 1, 3, 7, 4, 5, 2, 6 }, new int[] { 0, 0, 1, 2, 0, 0, 2, 1 }, new int[] { 0, 1, 6, 10, 4, 5, 3, 7, 8, 9, 2, 11 }, new int[] { 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0 });
        State moveB = new State(new int[] { 1, 5, 2, 3, 0, 4, 6, 7 }, new int[] { 1, 2, 0, 0, 2, 1, 0, 0 }, new int[] { 4, 8, 2, 3, 1, 5, 6, 7, 0, 9, 10, 11 }, new int[] { 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 });

        moves = new HashMap<String, State>();
        moves.put("U",  moveU);
        moves.put("U2", moveU.pow(2));
        moves.put("U'", moveU.pow(3));
        moves.put("D",  moveD);
        moves.put("D2", moveD.pow(2));
        moves.put("D'", moveD.pow(3));
        moves.put("L",  moveL);
        moves.put("L2", moveL.pow(2));
        moves.put("L'", moveL.pow(3));
        moves.put("R",  moveR);
        moves.put("R2", moveR.pow(2));
        moves.put("R'", moveR.pow(3));
        moves.put("F",  moveF);
        moves.put("F2", moveF.pow(2));
        moves.put("F'", moveF.pow(3));
        moves.put("B",  moveB);
        moves.put("B2", moveB.pow(2));
        moves.put("B'", moveB.pow(3));
    }
}
